/*
 Copyright (c) Sybase, Inc. 2012 All rights reserved. 
 
 In addition to the license terms set out in the Sybase License Agreement for 
 the Sybase Unwired Platform ("Program"), the following additional or different 
 rights and accompanying obligations and restrictions shall apply to the source 
 code in this file ("Code"). Sybase grants you a limited, non-exclusive, 
 non-transferable, revocable license to use, reproduce, and modify the Code 
 solely for purposes of (i) maintaining the Code as reference material to better 
 understand the operation of the Program, and (ii) development and testing of 
 applications created in connection with your licensed use of the Program. 
 The Code may not be transferred, sold, assigned, sublicensed or otherwise 
 conveyed (whether by operation of law or otherwise) to another party without 
 Sybase's prior written consent. The following provisions shall apply to any 
 modifications you make to the Code: (i) Sybase will not provide any maintenance 
 or support for modified Code or problems that result from use of modified Code; 
 (ii) Sybase expressly disclaims any warranties and conditions, express or 
 implied, relating to modified Code or any problems that result from use of the 
 modified Code; (iii) SYBASE SHALL NOT BE LIABLE FOR ANY LOSS OR DAMAGE RELATING 
 TO MODIFICATIONS MADE TO THE CODE OR FOR ANY DAMAGES RESULTING FROM USE OF THE 
 MODIFIED CODE, INCLUDING, WITHOUT LIMITATION, ANY INACCURACY OF DATA, LOSS OF 
 PROFITS OR DIRECT, INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, EVEN 
 IF SYBASE HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES; (iv) you agree 
 to indemnify, hold harmless, and defend Sybase from and against any claims or 
 lawsuits, including attorney's fees, that arise from or are related to the 
 modified Code or from use of the modified Code.
 */
package com.sybase.hwc;

import android.app.Activity;
import android.app.AlertDialog;
import android.app.ProgressDialog;
import android.content.DialogInterface;
import android.content.Intent;
import android.content.res.Configuration;
import android.graphics.Bitmap;
import android.net.Uri;
import android.os.Bundle;
import android.util.Log;
import android.view.Display;
import android.view.KeyEvent;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.view.ViewTreeObserver;
import android.view.WindowManager;
import android.webkit.GeolocationPermissions;
import android.webkit.JsResult;
import android.webkit.WebChromeClient;
import android.webkit.WebSettings;
import android.webkit.WebView;
import android.webkit.WebViewClient;

import com.sybase.hybridApp.amp.Consts;
import com.sybase.mo.MocaLog;

import org.apache.cordova.DroidGap;

/**
 * HybridAppContainer is the activity that displays the WebView hosting the
 * hybrid app.
 */
public class UiHybridAppContainer extends DroidGap
{

   private static final String LOG_TAG = "HybridAppContainer";

   // Receiver for challenges
   private ChallengeListener m_oChallengeListener = new ChallengeListener( this );

   private UiHybridAppController m_oHybridAppController = null;
   private WebView m_oWebView;
   private String m_sProgressText = "";

   private ViewTreeObserver.OnGlobalLayoutListener m_oLayoutListener;
   private boolean m_bOrientationChanged = false;

   private boolean m_bInitialized = false;

   // Helper to turn PhoneGap integration on/off
   private static final boolean USE_PHONEGAP = true;

   @Override
   public boolean onPrepareOptionsMenu( Menu oMenu )
   {
      m_oHybridAppController.onPrepareOptionsMenu( oMenu );

      return super.onPrepareOptionsMenu( oMenu );
   }

   @Override
   public boolean onOptionsItemSelected( MenuItem oItem )
   {
      m_oHybridAppController.onOptionsItemSelected( oItem );

      return super.onOptionsItemSelected( oItem );
   }

   @Override
   public void onCreate( Bundle savedInstanceState )
   {
      super.setBooleanProperty( "showTitle", true );
      super.onCreate( savedInstanceState );
   }

   @Override
   public void onResume()
   {
      super.onResume();

      // check to see if we need to redirect to the password entry or
      // change screen
      if ( PasswordPolicyMgr.redirect( this ) )
         return;

      if ( !m_bInitialized )
      {
         try
         {
            initWebView();
         } finally
         {
            m_bInitialized = true;
         }
      }

      m_oHybridAppController.onResume();

      // Register challenge listener
      HybridWebApplication.getInstance().setChallengeListener( m_oChallengeListener );
   }

   private void initWebView()
   {
      Intent oIntent = getIntent();

      // PhoneGap Change: Layout provided by PhoneGap
      if ( USE_PHONEGAP )
      {
         // PhoneGap Change: We must call DroidGap.init right away
         // so that the WebView is created
         super.init(); // <----- must comment this out if we no longer inherit
                       // from PhoneGap
         m_oWebView = this.appView;
      } else
      {
         setContentView( R.layout.hybridappcontainer );
         m_oWebView = ( WebView ) findViewById( R.id.webview );
      }

      int iStartMode = oIntent.getIntExtra( Consts.INTENT_PARAM_HYBRIDAPP_START_MODE,
               Consts.START_MODE_INVALID );

      if ( iStartMode == Consts.START_MODE_MESSAGE )
      {
         int iMessageID = oIntent.getIntExtra( Consts.INTENT_PARAM_HYBRIDAPP_MSG_ID, 0 );
         int iModuleID = oIntent.getIntExtra( Consts.INTENT_PARAM_HYBRIDAPP_MODULE_ID, 0 );
         int iModuleVersion = oIntent
                  .getIntExtra( Consts.INTENT_PARAM_HYBRIDAPP_MODULE_VERSION, -1 );
         m_sProgressText = oIntent.getStringExtra( Consts.INTENT_PARAM_HYBRIDAPP_PROGRESS_TEXT );

         m_oHybridAppController = new UiHybridAppController( this, m_oWebView, iMessageID,
                  iModuleID, iModuleVersion );
      } else if ( iStartMode == Consts.START_MODE_HYBRIDAPP )
      {
         int iHybridAppID = oIntent.getIntExtra( Consts.INTENT_PARAM_HYBRIDAPP_ID, 0 );
         m_sProgressText = oIntent.getStringExtra( Consts.INTENT_PARAM_HYBRIDAPP_PROGRESS_TEXT );

         m_oHybridAppController = new UiHybridAppController( this, m_oWebView, iHybridAppID );
      } else
      {
         setResult( Activity.RESULT_OK, null );
         finish();
         return;
      }

      // Check that controller was initialized. This could fail if the HybridApp
      // or message is missing
      if ( !m_oHybridAppController.isInitialized() )
      {
         setResult( Activity.RESULT_OK, null );
         finish();
         return;
      }

      // Register a layout changed listener to capture screen orientation
      // event(after orientation happens)
      View root = findViewById( android.R.id.content );
      ViewTreeObserver obs = root.getViewTreeObserver();
      m_oLayoutListener = new ViewTreeObserver.OnGlobalLayoutListener()
      {
         @Override
         public void onGlobalLayout()
         {
            if ( m_bOrientationChanged )
            {
               // call into javascript to notify an orientation change
               m_oWebView.loadUrl( "javascript:onFrameResize();" );
               m_bOrientationChanged = false;
            }
         }
      };
      obs.addOnGlobalLayoutListener( m_oLayoutListener );

      // Already set in the xml but due to Android bug in 2.2, must be
      // re-applied here.
      m_oWebView.setScrollBarStyle( View.SCROLLBARS_INSIDE_OVERLAY );

      WebSettings webSettings = m_oWebView.getSettings();
      webSettings.setSavePassword( false );
      webSettings.setSaveFormData( false );
      webSettings.setJavaScriptEnabled( true );

      m_oWebView.clearCache( true );

      // Setup localStorage
      webSettings.setDomStorageEnabled( true );
      AndroidHybridAppHelper.getInstance().setWebSettingsDatabasePath( webSettings,
               m_oHybridAppController );

      // Setup geolocation
      webSettings.setGeolocationEnabled( true );
      AndroidHybridAppHelper.getInstance().setWebSettingsGeolocationDatabasePath( webSettings,
               m_oHybridAppController );

      // PhoneGap Change: We use the PhoneGap version of these clients
      if ( !USE_PHONEGAP )
      {
         m_oWebView.setWebChromeClient( new HybridAppWebChromeClient() );
         m_oWebView.setWebViewClient( new HybridAppWebViewClient() );
      }

      String sBaseURL = m_oHybridAppController.getLaunchURL();

      // We can't directly load the url as some devices(Xoom) don't like having
      // a query string on the content provider url
      // Instead we provide the html data and url and let the content provider
      // handle any resources(js,css) that are linked in
      byte[] abData = m_oHybridAppController.getLaunchURLData();

      // PhoneGap Change: We must call through PhoneGap to load the URL
      if ( USE_PHONEGAP )
      {
         // PhoneGap takes cares of the loading spinner
         super.setStringProperty( "loadingDialog", m_sProgressText );

         // PhoneGap may timeout loading the web page
         super.setIntegerProperty( "loadUrlTimeoutValue", 300000 );

         // PhoneGap will load the URL
         super.loadUrlWithData( sBaseURL, abData );
      } else
      {
         m_oWebView.loadDataWithBaseURL( sBaseURL, new String( abData ), null, "utf-8", null );
      }
   }

   @Override
   protected void onPause()
   {
      // Unregister challenge listener
      HybridWebApplication.getInstance().setChallengeListener( null );

      if ( m_oHybridAppController != null )
         m_oHybridAppController.onPause();

      super.onPause();
   }

   @Override
   public void onDestroy()
   {
      super.onDestroy();

      View root = findViewById( android.R.id.content );
      ViewTreeObserver obs = root.getViewTreeObserver();
      obs.removeGlobalOnLayoutListener( m_oLayoutListener );

      if ( m_oWebView != null )
      {
         m_oWebView.clearCache( true );

         // Need to explicitly wipe the state of the webview
         // so the scriptable objects are GC
         m_oWebView.destroy();
      }

      if ( m_oHybridAppController != null )
         m_oHybridAppController.onDestroy();
   }

   @Override
   public void onUserInteraction()
   {
      // check to see if we need to redirect to the password entry or
      // change screen
      PasswordPolicyMgr.redirect( this );
      PasswordPolicyMgr.resetIdleCounter();
   }

   @Override
   public void onActivityResult( final int iRequest, final int iResult, final Intent oData )
   {
      // Process platform onActivityResult in another thread to avoid dead lock.
      Thread tRunner = new Thread( new Runnable()
      {
         @Override
         public void run()
         {
            processActivityResult( iRequest, iResult, oData );
         }
      } );
      tRunner.start();
   }

   /**
    * Process platform onActivityResult
    * 
    * @param iRequest
    * @param iResult
    * @param oData
    */
   private void processActivityResult( int iRequest, int iResult, Intent oData )
   {
      super.onActivityResult( iRequest, iResult, oData );
      m_oHybridAppController.onActivityResult( iRequest, iResult, oData );
   }

   @Override
   public boolean onKeyDown( int keyCode, KeyEvent event )
   {
      if ( m_oHybridAppController.onKeyDown( keyCode, event ) )
      {
         return true;
      }

      // Default behavior
      return super.onKeyDown( keyCode, event );
   }

   class HybridAppWebChromeClient extends WebChromeClient
   {
      public boolean onJsAlert( WebView view, String url, String message, final JsResult result )
      {
         new AlertDialog.Builder( UiHybridAppContainer.this ).setTitle( "HybridApp" )
                  .setMessage( message )
                  .setPositiveButton( android.R.string.ok, new AlertDialog.OnClickListener()
                  {
                     public void onClick( DialogInterface dialog, int which )
                     {
                        result.confirm();
                     }
                  } ).setCancelable( false ).create().show();
         return true;
      }

      public void onGeolocationPermissionsShowPrompt( String origin,
               GeolocationPermissions.Callback callback )
      {
         // Allow access to geolocation
         callback.invoke( origin, true, true );
      }
   }

   /**
    * HybridAppWebViewClient extends the WebViewClient The purpose of this class
    * is to provide any possible actions when the web page is loading. Currently
    * this displays the progress dialog when a page is being loaded
    * 
    */
   class HybridAppWebViewClient extends WebViewClient
   {
      private ProgressDialog m_oProgressDlg = null;

      /**
       * Shows a spinner on the screen
       */
      private void showProgressDialog( final String sMessage )
      {
         runOnUiThread( new Runnable()
         {
            @Override
            public void run()
            {
               if ( m_oProgressDlg == null )
               {
                  // Create progress dialog that cannot be canceled
                  m_oProgressDlg = ProgressDialog.show( UiHybridAppContainer.this, null, sMessage,
                           false );
               }
            }
         } );
      }

      /**
       * Hides the spinner on the screen
       */
      private void hideProgressDialog()
      {
         runOnUiThread( new Runnable()
         {
            @Override
            public void run()
            {
               try
               {
                  if ( m_oProgressDlg != null )
                  {

                     // when you dismiss the dialog the spinner stops and will
                     // not spin if you show it again
                     m_oProgressDlg.dismiss();
                     m_oProgressDlg = null;
                  }
               } catch ( Exception ex )
               {
                  MocaLog.getAmpHostLog().logMessage(
                           "Dismiss progress exception: " + ex.getMessage(),
                           MocaLog.eMocaLogLevel.Normal );
               }
            }
         } );
      }

      @Override
      public boolean shouldOverrideUrlLoading( WebView view, String url )
      {
         Log.d( LOG_TAG, "shouldOverrideUrlLoading" + url );

         // Local content we load within the browser. This may be an iframe.
         if ( url.indexOf( m_oHybridAppController.getProviderURL().toString() ) == 0 )
         {
            return false;
         }
         else
         {
            // We don't want to load any external links in our browser
            // Open these with an external viewer (mailto, tel, ...)
            Intent oIntent = new Intent( Intent.ACTION_VIEW, Uri.parse( url ) );
            startActivity( oIntent );

            return true;
         }
      }

      @Override
      public void onLoadResource( WebView view, String url )
      {
         Log.d( LOG_TAG, "onLoadResource" + url );
         MocaLog.getAmpHostLog().logMessage(
                  Consts.LOG_PREFIX_HYBRIDAPP_CONTAINER + "WebViewClient onloadresource: " + url,
                  MocaLog.eMocaLogLevel.Normal );
      }

      @Override
      public void onPageStarted( WebView view, String url, Bitmap favicon )
      {
         showProgressDialog( m_sProgressText );
         Log.d( LOG_TAG, "onPageStarted" + url );
         MocaLog.getAmpHostLog().logMessage(
                  Consts.LOG_PREFIX_HYBRIDAPP_CONTAINER + "WebViewClient onpagestarted: " + url,
                  MocaLog.eMocaLogLevel.Normal );

      }

      @Override
      public void onPageFinished( WebView view, String url )
      {
         hideProgressDialog();
         Log.d( LOG_TAG, "onPageFinished" + url );
         MocaLog.getAmpHostLog().logMessage(
                  Consts.LOG_PREFIX_HYBRIDAPP_CONTAINER + "WebViewClient onpageFinished: " + url,
                  MocaLog.eMocaLogLevel.Normal );

      }

      @Override
      public void onReceivedError( WebView view, int errorCode, String description,
               String failingUrl )
      {
         Log.d( LOG_TAG, "onReceivedError" + failingUrl );
         MocaLog.getAmpHostLog().logMessage(
                  Consts.LOG_PREFIX_HYBRIDAPP_CONTAINER + "WebViewClient onReceiveError: " +
                           failingUrl + ", description: " + description + ", errorCode: " +
                           errorCode, MocaLog.eMocaLogLevel.Normal );

      }
   }

   /**
    * This is a hack to let web client get the orientation change notification
    * after the orientation happens, ViewTreeObserver.OnGlobalLayoutListener is
    * hooked up in onCreate and will be triggered after orientation event.
    */
   @Override
   public void onConfigurationChanged( Configuration newConfig )
   {
      super.onConfigurationChanged( newConfig );

      Display display = ( ( WindowManager ) getSystemService( WINDOW_SERVICE ) )
               .getDefaultDisplay();
      int currOrientation = display.getOrientation();
      // check if the orientation is changed, filtering out other type of
      // changes
      if ( newConfig.orientation != currOrientation )
      {
         m_bOrientationChanged = true;
      }
   }
}
